/*
   Copyright (C) Cfengine AS

   This file is part of Cfengine 3 - written and maintained by Cfengine AS.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; version 3.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

  To the extent this program is licensed as part of the Enterprise
  versions of Cfengine, the applicable Commerical Open Source License
  (COSL) may apply to this file if you as a licensee so wish it. See
  included file COSL.txt.
*/

#include "cf3.defs.h"
#include "lastseen.h"

typedef struct
{
    double q;
    double expect;
    double var;
} QPoint0;

typedef struct
{
    char address[128];
    QPoint0 Q;
} KeyHostSeen0;

static bool LastseenMigrationVersion0(DBHandle *db)
{
    bool errors = false;
    DBCursor *cursor;

    if (!NewDBCursor(db, &cursor))
    {
        return false;
    }

    char *key;
    void *value;
    int ksize, vsize;

    while (NextDB(db, cursor, &key, &ksize, &value, &vsize))
    {
        if (ksize == 0)
        {
            CfOut(cf_inform, "", "LastseenMigrationVersion0: Database structure error -- zero-length key.");
            continue;
        }

        /* Only look for old [+-]kH -> IP<QPoint> entries */
        if (key[0] != '+' && key[0] != '-')
        {
            /* Warn about completely unexpected keys */

            if (key[0] != 'q' && key[0] != 'k' && key[0] != 'a')
            {
                CfOut(cf_inform, "", "LastseenMigrationVersion0: Malformed key found: %s", key);
            }

            continue;
        }

        bool incoming = key[0] == '-';
        const char *hostkey = key + 1;

        KeyHostSeen0 *old_data = value;

        /* Only migrate sane data */

        if (vsize != sizeof(KeyHostSeen0))
        {
            CfOut(cf_inform, "", "LastseenMigrationVersion0: invalid value size for key %s, entry is deleted",
                  key);
            DBCursorDeleteEntry(cursor);
            continue;
        }

        char hostkey_key[CF_BUFSIZE];
        snprintf(hostkey_key, CF_BUFSIZE, "k%s", hostkey);

        if (!WriteDB(db, hostkey_key, old_data->address, strlen(old_data->address) + 1))
        {
            CfOut(cf_inform, "", "Unable to write version 1 lastseen entry for %s", key);
            errors = true;
            continue;
        }

        char address_key[CF_BUFSIZE];
        snprintf(address_key, CF_BUFSIZE, "a%s", old_data->address);

        if (!WriteDB(db, address_key, hostkey, strlen(hostkey) + 1))
        {
            CfOut(cf_inform, "", "Unable to write version 1 reverse lastseen entry for %s", key);
            errors = true;
            continue;
        }

        char quality_key[CF_BUFSIZE];
        snprintf(quality_key, CF_BUFSIZE, "q%c%s", incoming ? 'i' : 'o', hostkey);

        /*
           Ignore malformed connection quality data
        */

        if (!isfinite(old_data->Q.q)
            || old_data->Q.q < 0
            || !isfinite(old_data->Q.expect)
            || !isfinite(old_data->Q.var))
        {
            CfOut(cf_inform, "", "Ignoring malformed connection quality data for %s", key);
            DBCursorDeleteEntry(cursor);
            continue;
        }

        KeyHostSeen data = {
            .lastseen = (time_t)old_data->Q.q,
            .Q = {
                /*
                   Previously .q wasn't stored in database, but was calculated
                   every time as a difference between previous timestamp and a
                   new timestamp. Given we don't have this information during
                   the database upgrade, just assume that last reading is an
                   average one.
                */
                .q = old_data->Q.expect,
                .dq = 0,
                .expect = old_data->Q.expect,
                .var = old_data->Q.var,
            }
        };

        if (!WriteDB(db, quality_key, &data, sizeof(data)))
        {
            CfOut(cf_inform, "", "Unable to write version 1 connection quality key for %s", key);
            errors = true;
            continue;
        }

        if (!DBCursorDeleteEntry(cursor))
        {
            CfOut(cf_inform, "", "Unable to delete version 0 lastseen entry for %s", key);
            errors = true;
        }
    }

    if (DeleteDBCursor(db, cursor) == false)
    {
        CfOut(cf_error, "", "LastseenMigrationVersion0: Unable to close cursor");
        errors = true;
    }

    if (!errors && !WriteDB(db, "version", "1", sizeof("1")))
    {
        errors = true;
    }

    return !errors;
}

bool LastseenMigration(DBHandle *db)
{
    char version[64];
    if (ReadDB(db, "version", version, sizeof(version)) == false)
    {
        return LastseenMigrationVersion0(db);
    }

    return true;
}
